#include <iomanip.h>
#include "AST.h"
#include "MethodDecl.h"
#include "UniqueId.h"
#include "compiler.h"
#include "decls.h"
#include "delimit.h"
#include "st-field.h"
#include "code-defs.h"
#include "code-util.h"
#include "cfg.h"


////////////////////////////////////////////////////////////
// 
// BASIC BLOCK ABSTRACTION
//
//    This code arranges expression-level CFG nodes into 
//    basic blocks, which speeds up dataflow calculations.
//
////////////////////////////////////////////////////////////

void setupBblocksMethod(TreeNode *t);

static int Bblock_number = 0;

Bblock::Bblock() {
  _preds = NULL;
  _succs = NULL;
  _cfglist = NULL;
  _defs = NULL;
  _kills = NULL;
  _defsOut = NULL;
  _defsIn = NULL;
  number = Bblock_number++;
  visited = 0;
  defsize = 0;
}

void Bblock::print(ostream &os) {
  os << endl << endl << "BBLOCK #" << number << ":\n";
  os << "PREDS:";
  ListIterator<Bblock *>bbIt = predIter();
  for (; !bbIt.isDone(); bbIt.next()) {
    Bblock *succ = *bbIt;
    os << " " << succ->number;
  }
  os << "\nSUCCS:";
  bbIt = succIter();
  for (; !bbIt.isDone(); bbIt.next()) {
    Bblock *succ = *bbIt;
    os << " " << succ->number;
  }
  os << endl;
  ListIterator<CfgNode *> cfgIt = cfgIter();
  for (; !cfgIt.isDone(); cfgIt.next()) {
    CfgNode *cfg = *cfgIt;
    os << endl << "<<<" << setbase (16) << "0x" << cfg->astNode() << ">>>" << setbase(10) << endl;
    cout << cfg->astNode()->oper_name() << endl;
    //    cfg->astNode()->print(os);
  }
}


Bblock *MethodDeclNode::getBblockRoot() {
  return bblockRoot;
}

void MethodDeclNode::setBblockRoot(Bblock *bblock) {
  bblockRoot = bblock;
}


// Calls setupBblocksMethod on all the MethodDeclNodes in the tree.
void setupBblocksFile(TreeNode *t) {
  const int childCount = t->arity();

  if (isMethodDeclNode(t)) {
    setupBblocksMethod(t);
  }
  else {
    for (int sweep = 0; sweep < childCount; sweep++) {
      // Need to quit the recursion when it's obvious there isn't
      // going to be a MethodDeclNode here.
      setupBblocksFile(t->child(sweep));
    }
  }
}

// Returns the statement that should get executed first in the given
// tree.
static StatementNode *findFirstStmt(TreeNode *t) {
  const int childCount = t->arity();
  TreeNode *first;

  if (t->isStatementNode() && !isBlockNode(t)) {
    return (StatementNode *)t;
  }

  for (int sweep = 0; sweep < childCount; sweep++) {
    first = findFirstStmt(t->child(sweep));
    if (first) {
      return (StatementNode *)first;
    }
  }
  return NULL;
}


// Creates the bblocks for a single method, and attaches the root to the
// bblockRoot field in the MethodDeclNode.
void setupBblocksMethod(TreeNode *t) {

  CfgNode *cfg;
  Bblock *firstBblock;
  
  assert(isMethodDeclNode(t));
  if (isOmittedNode(t->body())) {
    return;
  }
  if (isSynchronizedNode(t->body())) {
    cfg = t->body()->stmt()->getCfgExtent()->front();
  }
  else if (isBlockNode(t->body())) {
    cfg = t->body()->getCfgExtent()->front();
  }
  else {
    assert(0);
  }
  assert(cfg);
  firstBblock = new Bblock();
  ((MethodDeclNode *)t)->setBblockRoot(firstBblock);
  traverseCfgBblock(cfg, firstBblock, LastMethodNode(t));
  UnmarkBblocks(firstBblock);
}

// Does a DFO traversal of the CFG, collecting components with single
// predecessors and successors into single bblocks.
void traverseCfgBblock(CfgNode *c, Bblock *currBblock, CfgNode *lastCfg) {
  Bblock *newBblock;
  ListIterator<CfgNode *> succs;  

  assert(!c->bblock);
  currBblock->addCfg(c);
  if (c == lastCfg) {
    return;
  }
  succs = c->succIter();
  for (; !succs.isDone(); succs.next()) {
    CfgNode *child = *succs;
    if (child->bblock) {
      currBblock->addSucc(child->bblock);
      child->bblock->addPred(currBblock);
    }
    else if (c->numSucc > 1 || child->numPred > 1) {
      newBblock = new Bblock();
      traverseCfgBblock(child, newBblock, lastCfg);
      currBblock->addSucc(newBblock);
      newBblock->addPred(currBblock);
    }
    else {
      traverseCfgBblock(child, currBblock, lastCfg);
    }
  }
}

// Unmarks all the bblocks in the graph.
void UnmarkBblocks(Bblock *b) {
  b->nodeStatus(0);
  ListIterator<Bblock *> succs = b->succIter();
  for (; !succs.isDone(); succs.next()) {
    Bblock *child = *succs;
    if (child->nodeStatus()) {
      UnmarkBblocks(child);
    }
  }
}

static llist<Bblock *> *worklist;

// Recursion utility for setupWorklist().
void traverseRevPost(Bblock *b) {
  b->nodeStatus(1);
  ListIterator<Bblock *>succs = b->succIter();
  for (; !succs.isDone(); succs.next()) {
    Bblock *child = *succs;
    if (!child->nodeStatus()) {
      traverseRevPost(child);
    }
  }
  worklist = cons(b, worklist);
}

// Sets up the global worklist in reverse postorder.
void setupWorklist(Bblock *b) {
  UnmarkBblocks(b);
  worklist = NULL;
  traverseRevPost(b);
}
  

// Calls "action" on each bblock in the graph until nothing changes.
// "action" should return true if it changed the bblock, false otherwise.
void TraverseBblocksIter(Bblock *b, bool action(Bblock *b))
{
  setupWorklist(b);
  while (worklist) {
    bool changed = action(worklist->front());
    Bblock *current = worklist->front();
    current->nodeStatus(0);
    worklist = worklist->free();
    ListIterator<Bblock *>succs = current->succIter();
    for (; !succs.isDone(); succs.next()) {
      Bblock *child = *succs;
      if (!child->nodeStatus() && changed) {
	child->nodeStatus(1); // 1 means it's already on the list
	worklist = extend(worklist, cons(child));
      }
    }
  }
}

// Recursion utility for TraverseBblocks().
static void dfoBblocks(Bblock *b, void action(Bblock *b)) {
  b->nodeStatus(1);
  action(b);
  ListIterator<Bblock *> succs = b->succIter();
  for (; !succs.isDone(); succs.next()) {
    Bblock *child = *succs;
    if (!child->nodeStatus()) {
      dfoBblocks(child, action);
    }
  }
}

// Calls "action" on each bblock once, in DFO. 
void TraverseBblocks(Bblock *b, void action(Bblock *b)) {
  UnmarkBblocks(b);
  dfoBblocks(b, action);
}

// Unmarks the extent of the CFG which lies between "from" and "to", 
// inclusive.
void UnmarkCfgExts(CfgNode *from, CfgNode *to) {
  from->nodeStatus(0);
  if (from != to) {
    ListIterator<CfgNode *> succs = from->succIter();
    for (; !succs.isDone(); succs.next()) {
      CfgNode *child = *succs;
      if (child->nodeStatus()) {
	UnmarkCfgExts(child, to);
      }
    }
  }
}
  
// Traverses through the CF nodes between the given extents, calling "action"
// on each node. Call UnmarkCfgExts() before calling this routine.
void TraverseCfgExtents(CfgNode *from, CfgNode *to, void action(CfgNode *)) {
  from->nodeStatus(1);
  action(from);
  if (from != to) {
    ListIterator<CfgNode *> succs = from->succIter();
    for (; !succs.isDone(); succs.next()) {
      CfgNode *child = *succs;
      if (!child->nodeStatus()) {
	TraverseCfgExtents(child, to, action);
      }
    }
  }    
}

// Traverses through the CFG rooted at "c", calling "action" on each node.
void TraverseCfg(CfgNode *c, void action(CfgNode *))
{
    c->nodeStatus(1);
    action(c);
    ListIterator<CfgNode *> succs = c->succIter();
    for (; !succs.isDone(); succs.next()) {
	CfgNode *t = *succs;
	if (!t->nodeStatus())
	    TraverseCfg(t,action);
    }
}


///////////////////////////////////////////
//                                       //
// EXPRESSION-LEVEL CONTROL FLOW GRAPH   //
//                                       //
///////////////////////////////////////////

      
llist<CfgNode *> *TreeNode::getCfgExtent()
{
    undefined( "getCfgExtent" );
    return NULL;
}

llist<CfgNode *> *StatementNode::getCfgExtent()
{
  if (isSynchronizedNode(this)) {
    assert(0);
  }
  return cons(_cfgBeg, cons(_cfgEnd));
}

CfgNode *TreeNode::getContinueTarget()
{
    undefined( "getContinueTarget" );
    return NULL;
}

CfgNode *IterationNode::getContinueTarget()
{
    return _cfgContinueNode;
}

llist<CfgNode *> *TreeNode::getExceptionCfg()
{
    undefined( "getExceptionCfg" );
    return NULL;
}

llist<CfgNode *> *TryStmtNode::getExceptionCfg()
{
    return cons(_cfgCatch, cons(_cfgFinally));
}

// A post-order traversal of the nodes
llist<CfgNode *> *post_generic(TreeNode *n)
{
    const int childCount = n->arity();

    if (childCount == 0) {
	CfgNode *t = new CfgNode(n);
	return cons(t, cons(t));
    }
    
    llist<CfgNode *> *c = n->child(0)->makeCfg();
    CfgNode *c_beg = c->front();
    CfgNode *c_end = c->tail()->front();
    for (int sweep = 1; sweep < childCount; sweep++) {
	llist<CfgNode *> *next = n->child(sweep)->makeCfg();
	CfgNode *next_beg = next->front();
	CfgNode *next_end = next->tail()->front();
	c_end->addSucc(next_beg);
	next_beg->addPred(c_end);
	c_end = next_end;
    }
    CfgNode *t = new CfgNode(n);
    c_end->addSucc(t);
    t->addPred(c_end);
    return cons(c_beg, cons(t));
}

llist<CfgNode *> *append_cfg(llist<CfgNode *> *c1, llist<CfgNode *> *c2)
{
    CfgNode *c1_beg = c1->front();
    CfgNode *c1_end = c1->tail()->front();
    CfgNode *c2_beg = c2->front();
    CfgNode *c2_end = c2->tail()->front();
    c1_end->addSucc(c2_beg);
    c2_beg->addPred(c1_end);
    return cons(c1_beg, cons(c2_end));
}

llist<CfgNode *> *append_cfg(CfgNode *c1, llist<CfgNode *> *c2)
{
    CfgNode *c2_beg = c2->front();
    CfgNode *c2_end = c2->tail()->front();
    c1->addSucc(c2_beg);
    c2_beg->addPred(c1);
    return cons(c1, cons(c2_end));
}

llist<CfgNode *> *append_cfg(CfgNode *c1, CfgNode *c2)
{
    c1->addSucc(c2);
    c2->addPred(c1);
    return cons(c1, cons(c2));
}

llist<CfgNode *> *append_cfg(llist<CfgNode *> *c1, CfgNode *c2)
{
    CfgNode *c1_beg = c1->front();
    CfgNode *c1_end = c1->tail()->front();
    c1_end->addSucc(c2);
    c2->addPred(c1_end);
    return cons(c1_beg, cons(c2));
}

llist<CfgNode *> *explicit_1(TreeNode *node, TreeNode *child)
{
    CfgNode *t = new CfgNode(node);
    llist<CfgNode *> *c1 = cons(t, cons(t));
    llist<CfgNode *> *c2 = child->makeCfg();
    return append_cfg(c2, c1);
}

llist<CfgNode *> *explicit_2(TreeNode *node, TreeNode *child1, TreeNode *child2)
{
    CfgNode *t = new CfgNode(node);
    llist<CfgNode *> *c1 = cons(t, cons(t));
    llist<CfgNode *> *c2 = child1->makeCfg();
    llist<CfgNode *> *c3 = child2->makeCfg();
    return append_cfg(append_cfg(c2, c3), c1);
}

llist<CfgNode *> *explicit_3(TreeNode *node, TreeNode *child1,
			     TreeNode *child2, TreeNode *child3)
{
    CfgNode *t = new CfgNode(node);
    llist<CfgNode *> *c1 = cons(t, cons(t));
    llist<CfgNode *> *c2 = child1->makeCfg();
    llist<CfgNode *> *c3 = child2->makeCfg();
    llist<CfgNode *> *c4 = child3->makeCfg();
    return append_cfg(append_cfg(c2, append_cfg(c3, c4)), c1);
}

llist<CfgNode *> *TreeNode::makeCfg()
{
    return post_generic(this);
}

llist<CfgNode *> *ContinueNode::makeCfg()
{
    _cfgBeg = new CfgNode(this);
    _cfgEnd = new CfgNode(TreeNode::omitted);
    CfgNode *target = destination()->getContinueTarget();
    append_cfg(_cfgBeg, target);
    return cons(_cfgBeg, cons(_cfgEnd));
}

llist<CfgNode *> *ExpressionStmtNode::makeCfg()
{
    _cfgBeg = new CfgNode(TreeNode::omitted);
    _cfgEnd = new CfgNode(this);
    llist<CfgNode *> *c = expr()->makeCfg();
    return append_cfg(_cfgBeg, append_cfg(c, _cfgEnd));
}

llist<CfgNode *> *VarDeclNode::makeCfg()
{
  return explicit_1(this, initExpr());
}

CfgNode *LastMethodNode(TreeNode *p)
{
    while (  (strcmp(p->oper_name(), "MethodDeclNode") != 0)
	   && (strcmp(p->oper_name(), "ConstructorDeclNode") != 0))
	p = p->parent();
    llist<CfgNode *> *t;
    if (isSynchronizedNode(p->body())) {
      t = p->body()->stmt()->getCfgExtent();
    }
    else {
      t = p->body()->getCfgExtent();
    }
    CfgNode *dest = t->tail()->front();
    return dest;
}

llist<CfgNode *> *ReturnNode::makeCfg()
{
    _cfgBeg = new CfgNode(this);
    _cfgEnd = new CfgNode(TreeNode::omitted);
    CfgNode *dest = LastMethodNode(this);
    append_cfg(_cfgBeg, dest);
    return cons(_cfgBeg, cons(_cfgEnd));
}

void UnmarkCfg(CfgNode *c)
{
    c->nodeStatus(0);
    ListIterator<CfgNode *> succs = c->succIter();
    for (; !succs.isDone(); succs.next()) {
	CfgNode *t = *succs;
	if (t->nodeStatus())
	    UnmarkCfg(t);
    }
}


llist<CfgNode *> *MethodDeclNode::makeCfg()
{
    return body()->makeCfg();
}

llist<CfgNode *> *ConstructorDeclNode::makeCfg()
{
    llist<CfgNode *> *b = body()->makeCfg();
    llist<CfgNode *> *c = constructorCall()->makeCfg();
    return append_cfg(c, b);
}

llist<CfgNode *> *FieldDeclNode::makeCfg()
{
    CfgNode *t = new CfgNode(this);
    return cons(t, cons(t));
}

llist<CfgNode *> *ObjectFieldAccessNode::makeCfg()
{
    return explicit_1(this, object());
}

llist<CfgNode *> *TypeFieldAccessNode::makeCfg()
{
    CfgNode *t = new CfgNode(this);
    return cons(t, cons(t));
}

llist<CfgNode *> *ThisFieldAccessNode::makeCfg()
{
    CfgNode *t = new CfgNode(this);
    return cons(t, cons(t));
}

llist<CfgNode *> *SuperFieldAccessNode::makeCfg()
{
    CfgNode *t = new CfgNode(this);
    return cons(t, cons(t));
}

llist<CfgNode *> *EmptyStmtNode::makeCfg()
{
    CfgNode *t = new CfgNode(this);
    _cfgBeg = _cfgEnd = t;
    return cons(t, cons(t));
}

static CfgNode *ExceptionTarget(TreeNode *p)
{
    CfgNode *dest;
    TreeNode *par = p->parent();
    while (  (strcmp(par->oper_name(), "MethodDeclNode") != 0)
	     && (strcmp(par->oper_name(), "ConstructorDeclNode") != 0)
	     && (strcmp(par->oper_name(), "StaticInitNode") != 0) ) {
	if (strcmp(par->oper_name(), "TryStmtNode") == 0) {
	    if (strcmp(p->oper_name(), "TryNode") == 0)
		dest = par->getExceptionCfg()->front();
	    else
		dest = par->getExceptionCfg()->tail()->front();
	    return dest;
	}
	p = par;
	par = par->parent();
    }
    if (strcmp(par->oper_name(), "StaticInitNode") == 0)
	return NULL;
    if (isSynchronizedNode(par->body())) {
      dest = par->body()->stmt()->getCfgExtent()->tail()->front();
    }
    else {
      dest = par->body()->getCfgExtent()->tail()->front();
    }
    return dest;
}

llist<CfgNode *> *TryStmtNode::makeCfg()
{
    _cfgBeg = new CfgNode(TreeNode::omitted);
    _cfgEnd = new CfgNode(this);
    _cfgCatch = new CfgNode(TreeNode::omitted);
    _cfgFinally = new CfgNode(TreeNode::omitted);

    llist<CfgNode *> *c1 = block()->makeCfg();
    CfgNode *c1_beg = c1->front();
    CfgNode *c1_end = c1->tail()->front();
    append_cfg(_cfgBeg, c1_beg);
    append_cfg(c1_end, _cfgEnd);

    llist<CfgNode *> *c2 = catches()->makeCfg();
    CfgNode *c2_beg = c2->front();
    CfgNode *c2_end = c2->tail()->front();
    append_cfg(_cfgCatch, c2_beg);

    llist<CfgNode *> *c3 = finally()->makeCfg();
    CfgNode *c3_beg = c3->front();
    CfgNode *c3_end = c3->tail()->front();
    append_cfg(c1_end, c3_beg);
    append_cfg(c2_end, c3_beg);
    append_cfg(c3_end, _cfgEnd);
    append_cfg(_cfgFinally, c3_beg);

    CfgNode *c4 = ExceptionTarget(this);
    if (c4)
	append_cfg(c3_end, c4);
	
    return cons(_cfgBeg, cons(_cfgEnd));
}

llist<CfgNode *> *CatchListNode::makeCfg()
{
    CfgNode *catchBeg = new CfgNode(TreeNode::omitted);
    CfgNode *catchEnd = new CfgNode(this);

    const int childCount = arity();
    for (int sweep=0; sweep<childCount; sweep++) {
	llist<CfgNode *> *c = child(sweep)->makeCfg();
	CfgNode *c_beg = c->front();
	CfgNode *c_end = c->tail()->front();
	append_cfg(catchBeg, c_beg);
	append_cfg(c_end, catchEnd);
    }
	
    return cons(catchBeg, cons(catchEnd));
}

llist<CfgNode *> *ForEachStmtNode::makeCfg()
{
    _cfgBeg = new CfgNode(TreeNode::omitted);
    _cfgEnd = new CfgNode(this);
    llist<CfgNode *> *c1 = vars()->makeCfg();
    llist<CfgNode *> *c2 = stmt()->makeCfg();

    CfgNode *c1_beg = c1->front();
    CfgNode *c1_end = c1->tail()->front();
    CfgNode *c2_beg = c2->front();
    CfgNode *c2_end = c2->tail()->front();

    append_cfg(_cfgBeg, c1_beg);
    append_cfg(c1_end, c2_beg);
    append_cfg(c2_end, c1_beg);
    append_cfg(c1_end, _cfgEnd);
    return cons(_cfgBeg, cons(_cfgEnd));    
}

llist<CfgNode *> *BlockNode::makeCfg()
{
    _cfgBeg = new CfgNode(TreeNode::omitted);
    _cfgEnd = new CfgNode(this);
    llist<CfgNode *> *c = stmts()->makeCfg();
    return append_cfg(_cfgBeg, append_cfg(c, _cfgEnd));
}

llist<CfgNode *> *LabeledStmtNode::makeCfg()
{
    _cfgBeg = new CfgNode(TreeNode::omitted);
    _cfgEnd = new CfgNode(this);
    llist<CfgNode *> *t = stmt()->makeCfg();
    return append_cfg(_cfgBeg, append_cfg(t, _cfgEnd));
}

llist<CfgNode *> *IfStmtNode::makeCfg()
{
    _cfgBeg = new CfgNode(TreeNode::omitted);
    _cfgEnd = new CfgNode(this);

    llist<CfgNode *> *c1 = condition()->makeCfg();
    llist<CfgNode *> *c2 = thenPart()->makeCfg();
    llist<CfgNode *> *c3 = elsePart()->makeCfg();

    append_cfg(c1, c2);
    append_cfg(c1, c3);
    append_cfg(_cfgBeg, c1);
    append_cfg(c2, _cfgEnd);
    append_cfg(c3, _cfgEnd);
    return cons(_cfgBeg, cons(_cfgEnd));
}

llist<CfgNode *> *WhileNode::makeCfg()
{
    _cfgBeg = new CfgNode(TreeNode::omitted);
    _cfgEnd = new CfgNode(this);
    _cfgContinueNode = _cfgBeg;

    llist<CfgNode *> *c1 = test()->makeCfg();
    llist<CfgNode *> *c2 = stmt()->makeCfg();

    CfgNode *c1_beg = c1->front();
    CfgNode *c1_end = c1->tail()->front();
    CfgNode *c2_beg = c2->front();
    CfgNode *c2_end = c2->tail()->front();

    append_cfg(_cfgBeg, c1_beg);
    append_cfg(c1_end, c2_beg);
    append_cfg(c1_end, _cfgEnd);
    append_cfg(c2_end, _cfgBeg);
    return cons(_cfgBeg, cons(_cfgEnd));
}

llist<CfgNode *> *DoNode::makeCfg()
{
    _cfgBeg = new CfgNode(TreeNode::omitted);
    _cfgEnd = new CfgNode(this);

    llist<CfgNode *> *c1 = test()->makeCfg();
    CfgNode *c1_beg = c1->front();
    CfgNode *c1_end = c1->tail()->front();
    _cfgContinueNode = c1_beg;

    llist<CfgNode *> *c2 = stmt()->makeCfg();
    CfgNode *c2_beg = c2->front();
    CfgNode *c2_end = c2->tail()->front();

    append_cfg(_cfgBeg, c2_beg);
    append_cfg(c2_end, c1_beg);
    append_cfg(c1_end, c2_beg);
    append_cfg(c1_end, _cfgEnd);
    return cons(_cfgBeg, cons(_cfgEnd));
}

llist<CfgNode *> *ForNode::makeCfg()
{
    _cfgBeg = new CfgNode(TreeNode::omitted);
    _cfgEnd = new CfgNode(this);

    llist<CfgNode *> *c1 = init()->makeCfg();
    CfgNode *c1_beg = c1->front();
    CfgNode *c1_end = c1->tail()->front();
    append_cfg(_cfgBeg, c1_beg);

    llist<CfgNode *> *c2 = test()->makeCfg();
    CfgNode *c2_beg = c2->front();
    CfgNode *c2_end = c2->tail()->front();
    append_cfg(c1_end, c2_beg);
    append_cfg(c2_end, _cfgEnd);
    _cfgContinueNode = c2_beg;

    llist<CfgNode *> *c3 = stmt()->makeCfg();
    CfgNode *c3_beg = c3->front();
    CfgNode *c3_end = c3->tail()->front();
    append_cfg(c2_end, c3_beg);

    llist<CfgNode *> *c4 = update()->makeCfg();
    CfgNode *c4_beg = c4->front();
    CfgNode *c4_end = c4->tail()->front();
    append_cfg(c3_end, c4_beg);
    return cons(_cfgBeg, cons(_cfgEnd));
}

llist<CfgNode *> *SwitchNode::makeCfg()
{
    _cfgBeg = new CfgNode(TreeNode::omitted);
    _cfgEnd = new CfgNode(this);
    llist<CfgNode *> *c1 = expr()->makeCfg();
    append_cfg(_cfgBeg, c1);

    TreeListNode *swB = switchBlocks();
    int c = swB->arity();
    CfgNode *prev = NULL;
    for (int i=0; i<c; i++) {
	llist<CfgNode *> *t = swB->child(i)->makeCfg();
	append_cfg(c1, t);
	if (i != 0) {
	    append_cfg(prev, t);
	}
	prev = t->tail()->front();
    }
    if (prev != NULL)
	append_cfg(prev, _cfgEnd);
    else
	append_cfg(c1, _cfgEnd);
    return cons(_cfgBeg, cons(_cfgEnd));
}

llist<CfgNode *> *BreakNode::makeCfg()
{
    _cfgBeg = new CfgNode(this);
    _cfgEnd = new CfgNode(TreeNode::omitted);
    llist<CfgNode *> *t = destination()->getCfgExtent();
    CfgNode *target = t->tail()->front();
    append_cfg(_cfgBeg, target);
    return cons(_cfgBeg, cons(_cfgEnd));
}

llist<CfgNode *> *ThrowNode::makeCfg()
{
    _cfgBeg = new CfgNode(this);
    _cfgEnd = new CfgNode(TreeNode::omitted);

    llist<CfgNode *> *c = expr()->makeCfg();
    CfgNode *c_beg = c->front();
    CfgNode *c_end = c->tail()->front();
    append_cfg(_cfgBeg, c_beg);
    
    CfgNode *dest = ExceptionTarget(this);
    assert(dest);
    append_cfg(c_end, dest);
    return cons(_cfgBeg, cons(_cfgEnd));    
}

llist<CfgNode *> *ExprNode::makeCfg()
{
    return post_generic(this);
}

llist<CfgNode *> *IfExprNode::makeCfg()
{
    return post_generic(this);
}

llist<CfgNode *> *MethodCallNode::makeCfg()
{
    MethodDecl *d = (MethodDecl *)method()->simpName()->decl();
    TreeNode *declNode = d->source();
    if (!isOmittedNode(declNode) && declNode->throws()->arity()) {
	llist<CfgNode *> *c = post_generic(this);
	CfgNode *c_beg = c->front();
	CfgNode *c_end = c->tail()->front();
	CfgNode *c_exc = ExceptionTarget(this);
	append_cfg(c_end, c_exc);
	return c;
    } else
	return post_generic(this);
}

llist<CfgNode *> *AllocateNode::makeCfg()
{
    return post_generic(this);
}

llist<CfgNode *> *AllocateArrayNode::makeCfg()
{
    return post_generic(this);
}

llist<CfgNode *> *BinaryArithAssignNode::makeCfg()
{
    return post_generic(this);
}

llist<CfgNode *> *ShiftAssignNode::makeCfg()
{
    return post_generic(this);
}

llist<CfgNode *> *AssignNode::makeCfg()
{
    return post_generic(this);
}

llist<CfgNode *> *InstanceOfNode::makeCfg()
{
    return explicit_1(this, opnd0());
}

llist<CfgNode *> *NameNode::makeCfg()
{
    CfgNode *t = new CfgNode(this);
    return cons(t, cons(t));
}
