#include "AST.h"
#include "CodeContext.h"
#include "code.h"
#include "lgMacro.h"
#include "decls.h"
#include "code-assign.h"

const string assignPrefix(AssignKind kind, bool isStatic)
{
  static char *assignMacros[][2] = 
  {
    "",   "",
    "PL", "SPL",
    "PG", "SPG"
  };

  return assignMacros[kind][isStatic];
}

const string assignMacro(AssignKind kind, bool isStatic, bool isLocal)
{
  return assignPrefix(kind, isStatic) + lgMacro("ASSIGN", isLocal);
}

AssignKind assignKind(const TypeNode *type)
{
  if (type->isReference())
    if (type->isLocal()) return assignLocal;
    else return assignGlobal;
  else return assignNormal;
}

const string makeAssign(const string &lvalue, const string &value,
			const bool isLocal, AssignKind kind, bool isStatic)
{
  return assignMacro(kind, isStatic, isLocal)
    + '(' + lvalue + "," +  value + ")";
}


const string TreeNode::receiveAssign(const string &, const string &, AssignKind)
{
  undefined("receiveAssign");
  return "";
}


const string ObjectFieldAccessNode::receiveAssign(const string &lvalue,
						  const string &value,
						  AssignKind kind)
{
  bool isStatic = (decl()->modifiers() & Static) != 0;
  return makeAssign(lvalue, value, isLocalLvalue(), kind, isStatic);
}


const string TypeFieldAccessNode::receiveAssign(const string &lvalue,
						const string &value,
						AssignKind kind)
{
  return makeAssign(lvalue, value, true, kind, true);
}


const string ExprNode::emitAssignBinOp(CodeContext &context,
					const char op[])
{
    if (opnd0()->isSimpleLvalue()) { /* a local variable */
        const string prior = opnd0()->defineTemporary(context);
	const string modifier = opnd1()->emitExpression(context);
	context << opnd0()->emitExpression(context) << " = "
		<< prior + ' ' + op + " (" + modifier + ");" << endCline;
	return opnd0()->emitExpression(context);

    } else {
	const string lval = opnd0()->getLvalue(context);
	const string prior = opnd0()->declareTemporary(context);
	context << lgMacro("DEREF", opnd0()->isLocalLvalue())
		<< '(' << prior << ", " << lval << ");" << endCline;
	const string modifier = opnd1()->emitExpression(context);
	const string value = buildTemporary( context, '(' + prior + ") "
					     + op + " (" + modifier + ')' );
	context << opnd0()->receiveAssign(lval, value, assignNormal)
		<< ";" << endCline;
	return value;
    }
}

// True for non-static ObjectFieldAccessNode's whose object is immutable
bool TreeNode::isFieldOfImmutable() const
{
  return false;
}
  
bool FieldAccessNode::isFieldOfImmutable() const
{
  return !(decl()->modifiers() & Static) &&
    accessedObjectType()->isImmutable();
}
  
const string AssignNode::emitExpression2(CodeContext &context)
{
    if (opnd0()->isSimpleLvalue()) {
        return opnd0()->emitExpression(context) + " = "
	  + opnd1()->emitExpression(context);

    } else if (opnd0()->isFieldOfImmutable()) {
	const string lval = opnd0()->getLvalue(context);
	const string value = opnd1()->simpleVar(context);
	context << "*(" << lval << ") = " << value << ";" << endCline;
	return value;
    } else {
	const AssignKind kind = assignKind(type());
	const string lval = opnd0()->getLvalue(context);
	const string value = opnd1()->simpleVar(context);
	context << opnd0()->receiveAssign(lval, value, kind) << ";" << endCline;
	return value;
    }
}


const string MultAssignNode::emitExpression2(CodeContext &context)
{
    return emitAssignBinOp(context, "*");
}


const string DivAssignNode::emitExpression2(CodeContext &context)
{
    return emitAssignBinOp(context, "/");
}


const string RemAssignNode::emitExpression2(CodeContext &context)
{
    return emitAssignBinOp(context, "%");
}


const string PlusAssignNode::emitExpression2(CodeContext &context)
{
    return emitAssignBinOp(context, "+");
}


const string MinusAssignNode::emitExpression2(CodeContext &context)
{
    return emitAssignBinOp(context, "-");
}


const string LeftShiftLogAssignNode::emitExpression2(CodeContext &context)
{
    return emitAssignBinOp(context, "<<");
}


const string RightShiftLogAssignNode::emitExpression2(CodeContext &context)
{
    return emitAssignBinOp(context, ">>");
}


const string RightShiftArithAssignNode::emitExpression2(CodeContext &context)
{
    return emitAssignBinOp(context, ">>");
}


const string BitAndAssignNode::emitExpression2(CodeContext &context)
{
    return emitAssignBinOp(context, "&");
}


const string BitXorAssignNode::emitExpression2(CodeContext &context)
{
    return emitAssignBinOp(context, "^");
}


const string BitOrAssignNode::emitExpression2(CodeContext &context)
{
    return emitAssignBinOp(context, "|");
}
