#define __SINST__		// g++ header bug
#include <map>
#include <strstream.h>
#include <utility>
#include "AST.h"
#include "ClassDecl.h"
#include "CodeContext.h"
#include "CtPointInstance.h"
#include "CtReference.h"
#include "MethodDecl.h"
#include "casts.h"
#include "code.h"
#include "code-call.h"
#include "config.h"
#include "ctBox.h"
#include "decls.h"
#include "delimit.h"
#include "domain-decls.h"
#include "jni-bridge.h"
#include "code-grid.h"
#include "code-util.h"
#include "decimalString.h"
#include "lgMacro.h"

bool opt_finalize = true;

typedef map< string, string, less<string> > stringmap;

static const char *retTypeSig(string sig)
{
    int i;

    for (i=0; sig[i] != ')'; i++);
    switch (sig[i+1]) {
      // case 'T':
    case 'V': return "";
    case 'Q':
    case 'L': return ".l";
    case 'Z': return ".z";
    case 'B': return ".b";
    case 'C': return ".c";
    case 'S': return ".s";
    case 'I': return ".i";
    case 'J': return ".j";
    case 'F': return ".f";
    case 'D': return ".d";
    default:  return 0;		// should never happen
    }
}


static void pushThis( const string &instance, const MethodDecl &method,
		      vector< string > &parameters, const TypeNode &objectType,
		      bool alwaysCast )
{
    bool isRef = objectType.isReference();
    if (alwaysCast || !(method.modifiers() & Common::Static))
        if (alwaysCast || isRef) {
	    const bool thisLocal = method.modifiers() & Common::Local;
	    const CtType &rawType = method.container()->cType();
	    const CtType &typeName = ctBox( rawType, thisLocal );
	    if (alwaysCast || typeName != objectType.cType()) {
	        parameters.push_back("*(" + typeName + "*) &" + instance);
	        return;
	    }
	}

    /* Default: no cast is needed. */
    parameters.push_back(instance);
}


bool ObjectFieldAccessNode::canCallStatically() const
{
  assert( decl()->category() == Decl::Method );
  const TypeDecl &receiver = *accessedObjectType()->decl();

  if (receiver.category() == Decl::Class)
    {      
      // fundamentally cannot be overridden
      if (decl()->modifiers() & (Final | Private | Static))
	return true;

      // dispatched to a final class
      if (receiver.modifiers() & Final)
	return true;
      
      // not observed to be overridden anywhere
      if (codeGen_main && opt_finalize && decl()->overriders()->empty())
	return true;
    }

  return false;
}


static const string dispatch( CodeContext &context,
			      const ObjectFieldAccessNode &access,
			      const string &instance,
			      vector< string > &parameters )
{
    const TypeNode &objectType = *access.accessedObjectType();
    ClassDecl &classDecl = static_cast< ClassDecl & >(*objectType.decl());    
    const MethodDecl &methodDecl = static_cast< MethodDecl & >( *access.decl() );
    assert( classDecl.category() & (Decl::Class | Decl::Interface) );
    assert( methodDecl.category() == Decl::Method );
    
    const bool isLocal = objectType.isLocal();
    const bool needsDispatch = !access.canCallStatically();

    if (objectType.isReference())
	context << lgMacro("CHECK_NULL", isLocal )
		<< '(' << instance << ')' << ';'
		<< endCline;

    pushThis( instance, methodDecl, parameters, objectType, false );

    if (needsDispatch) {
	if (methodDecl.container()->modifiers() & Common::Interface)
	    return "IFACE_DISPATCH_"
		+ methodDecl.cMethodNameStatic()
		+ '(' + instance + ')';
	else {
	    const string descTemp = getClassInfo( context, classDecl,
						  instance, isLocal );
	    return descTemp + "->" + methodDecl.cMethodNameDynamic();
	}
    } else
	return methodDecl.cMethodNameStatic();
}


const string MethodCallNode::emitExpression2( CodeContext &context )
{
    Decl *d = method()->simpName()->decl();

    // Here are the cases that we have to worry about:
    // The method call can either be on a java-defined class or titanium-defined class.
    // The call can be either a static method call or instance method call.
    // The object (for an instance method call) can be either "this" or an
    // explicitly defined object.

    bool useJni = false;
    if (d->modifiers() & Common::Static) {
      if (!isThisFAN(method()) && !isSFAN(method())) {
	TreeNode *t;
	if (isTypeFAN(method()))
	  t = method()->ftype()->decl()->asType();
	else
	  t = method()->object()->type();
	useJni = javaDefinedType(t);
      }
    } else {
      if (!isThisFAN(method()) && !isSFAN(method())) {
	TreeNode *t = method()->object()->type();
	useJni = javaDefinedType(t);
      }
    }

    if (useJni) {
	vector< string > parameters;
	string methodName;
	
	if (d->modifiers() & Common::Static) {
	    // Ti_CallJavaStaticMethod(cl_name, mt_name, signature, args).ret
	    methodName = "Ti_CallJavaStaticMethod";
	    parameters.push_back( '"' + javaClassName(d->container()) + '"' );
	} else {
	    // Ti_CallJavaMethod(obj, mt_name, signature, args).ret
	    methodName = "Ti_CallJavaMethod";
	    parameters.push_back( method()->object()->emitExpression( context ) );
	}
	parameters.push_back( '"' + *(d->name()) + '"' );
	parameters.push_back( '"' + d->type()->signature() + '"' );

	args()->emitExpressionList( context, parameters );
	return methodName + '(' + delimit( parameters ) + ')'
	    + retTypeSig(d->type()->signature());

    } else
	return method()->emitMethodCall( context, *args() );
}

static const string pointDispatch(Decl * const methodDecl, int arity)
{
  TypeListNode *params = methodDecl->type()->paramTypes();
  string conv_name;
  const string name = *methodDecl->name();

  string ending;
  for (int i=0; i<params->arity(); i++) {
    TypeNode &paramType = *params->child(i);

    assert(!paramType.isJavaArrayType());
    assert(!paramType.isTitaniumArrayType());

    if (paramType.hasFormalSignature()) {
      ending += paramType.formalSignature();
    } else {
      ending += paramType.cType();
    }
  }

  static stringmap m;
  static bool inited = false;

  if (!inited) {
    static char *s[] = { "+", "add",
			 "-", "sub",
			 "*", "mul",
			 "/", "div",
			 "==", "isEqual",
			 "!=", "isNotEqual",
			 "<", "isLessThan",
			 "<=", "isLessThanEqual",
			 ">", "isGreaterThan",
			 ">=", "isGreaterThanEqual",
			 "set", "set",
			 "get", "get",
			 "getPrevious", "getPrevious",
			 "setThis", "setThis",
			 "setAllThis", "setAllThis",
			 "permute", "permute",
			 "getLowerBound", "getLowerBound",
			 "getUpperBound", "getUpperBound",
			 /* "dot", "dot", */
			 "getLcm", "getLcm",
			 "all", "all",
			 "direction", "setDirection",
			 "toString", "toString",
			 "arity", "arity" };

    for (int i = 0; i < sizeof(s) / sizeof(s[0]); i += 2) {
      m[string(s[i])] = string(s[i + 1]);
      m[string(s[i + 1])] = string(s[i + 1]);
    }
    inited = true;
  }

  if ((conv_name = m[name]) == "") {
    printf("Internal Error: Unknown point call to \"%s\"\n", name.c_str());
    return string("???");
  }
  return string("m") + int2string(conv_name.length()) + conv_name +
    ending + "m" + PointNDecl[arity - 1]->cType().name;
}

string rectDomainMethodSlice(Decl * const methodDecl,
			     string instance,
			     int arity,
			     string dim_param,
			     CodeContext &os,
			     const string temp_p0,
			     const string temp_p1,
			     const string temp_stride)
{
  const CtType &pointFull   = *new CtPointInstance( arity );
  const CtType &pointSliced = *new CtPointInstance( arity - 1 );
  string arity_str;
  string arity_m1_str;

  arity_str = int2string(arity);
  arity_m1_str = int2string(arity-1);

  os.declare( temp_p0,     pointSliced );
  os.declare( temp_p1,     pointSliced );
  os.declare( temp_stride, pointSliced );

  os << "{" << endCline;
  os << "  P" << arity
     << " slicevar_old_p0, slicevar_old_p1, slicevar_old_stride;" << endCline;
  os << "  int slicevar_index, slicevar_count;" << endCline;
  os << "  int slicevar_kill_dim = (" << dim_param << ") - 1;" << endCline;
  os << "  slicevar_old_p0 = "
     << MANGLE_TI_DOMAINS_RECTDOMAIN_DISPATCH(<<, arity_str, "getLowerBound", "")
     << instance << ");" << endCline;
  os << "  slicevar_old_p1 = "
     << MANGLE_TI_DOMAINS_RECTDOMAIN_DISPATCH(<<, arity_str, "getUpperBound", "")
     << instance << ");" << endCline;
  os << "  slicevar_old_stride = "
     << MANGLE_TI_DOMAINS_RECTDOMAIN_DISPATCH(<<, arity_str, "getStride", "")
     << instance << ");" << endCline;

  os << "  for (slicevar_index = 0, slicevar_count = 0; slicevar_index < "
     << arity << "; slicevar_index++) {" << endCline;
  os << "    if (slicevar_index != slicevar_kill_dim) {" << endCline;
  os << "      " << temp_p0 << " = "
     << SET_POINT_INDEX_START(<<, arity_m1_str) << temp_p0 << ", "
     << SET_POINT_INDEX_END(<<, "slicevar_count", MANGLE_TI_DOMAINS_POINT_GET(<<, arity_str)
     << "slicevar_old_p0, slicevar_index)") << ";"
     << endCline;
  os << "      " << temp_p1 << " = "
     << SET_POINT_INDEX_START(<<, arity_m1_str) << temp_p1 << ", "
     << SET_POINT_INDEX_END(<<, "slicevar_count", MANGLE_TI_DOMAINS_POINT_GET(<<, arity_str)
     << "slicevar_old_p1, slicevar_index)") << ";"
     << endCline;
  os << "      " << temp_stride << " = "
     << SET_POINT_INDEX_START(<<, arity_m1_str) << temp_stride
     << ", "
     << SET_POINT_INDEX_END(<<, "slicevar_count", MANGLE_TI_DOMAINS_POINT_GET(<<, arity_str)
     << "slicevar_old_stride, slicevar_index)") << ";"
     << endCline;
  os << "      slicevar_count++;" << endCline;
  os << "    }" << endCline;
  os << "  }" << endCline;
  os << "}" << endCline;

  return (string) NEW_RECTDOMAIN_3PT_UPB_START(+, arity_m1_str) +
    temp_p0 +
    "," +
    temp_p1 +
    "," +
    temp_stride +
    NEW_RECTDOMAIN_3PT_UPB_END();
}

static const string domainDispatch(Decl * const methodDecl, int arity, bool isDomain, bool genStatic)
{
  TypeListNode *params = methodDecl->type()->paramTypes();
  string conv_name;
  string ending;
  const char* name = methodDecl->name()->c_str();
  string result;

  for (int i=0; i<params->arity(); i++) {
    TypeNode &paramType = *params->child(i);

    if (paramType.hasFormalSignature()) {
      ending += paramType.formalSignature();
    } else {
      ending += paramType.cType();
    }
  }

  if (strcmp(name, "isRectangular") == 0) {
    conv_name = "isRectangular";
  } else if (strcmp(name, "+") == 0) {
    conv_name = "add";
  } else if (strcmp(name, "-") == 0) {
    conv_name = "difference";
  } else if (strcmp(name, "*") == 0) {
    conv_name = "multiply";
  } else if (strcmp(name, "/") == 0) {
    conv_name = "divide";
  } else if (strcmp(name, "==") == 0) {
    conv_name = "isEqual";
  } else if (strcmp(name, "!=") == 0) {
    conv_name = "isNotEqual";
  } else if (strcmp(name, "<") == 0) {
    conv_name = "isStrictSubset";
  } else if (strcmp(name, "<=") == 0) {
    conv_name = "isSubset";
  } else if (strcmp(name, ">") == 0) {
    conv_name = "isStrictSuperset";
  } else if (strcmp(name, ">=") == 0) {
    conv_name = "isSuperset";
  } else if (strcmp(name, "min") == 0) {
    conv_name = "getLowerBound";
  } else if (strcmp(name, "max") == 0) {
    conv_name = "getUpperMaxPoint";
  } else if (strcmp(name, "lwb") == 0) {
    conv_name = "getLowerBound";
  } else if (strcmp(name, "upb") == 0) {
    conv_name = "getUpperBound";
  } else if (strcmp(name, "stride") == 0) {
    conv_name = "getLoopingStride";
  } else if (strcmp(name, "accrete") == 0) {
    conv_name = "accrete";
  } else if (strcmp(name, "shrink") == 0) {
    conv_name = "shrink";
  } else if (strcmp(name, "border") == 0) {
    conv_name = "border";
  } else if (strcmp(name, "contains") == 0) {
    conv_name = "doesContain";
  } else if (strcmp(name, "boundingBox") == 0) {
    conv_name = "getBoundingBox";
  } else if (strcmp(name, "arity") == 0) {
    conv_name = "getArity";
  } else if (strcmp(name, "size") == 0) {
    conv_name = "getNumPoints";
  } else if (strcmp(name, "isNull") == 0) {
    conv_name = "isNull";
  } else if (strcmp(name, "toString") == 0) {
    conv_name = "toString";
  } else if (strcmp(name, "permute") == 0) {
    conv_name = "permute";
  } else if (isDomain && (strcmp(name, "RectDomainList") == 0)) {
    conv_name = "getRectDomains";
  } else if (isDomain && (strcmp(name, "PointList") == 0)) {
    conv_name = "getPoints";
  } else if (isDomain && (strcmp(name, "toDomain") == 0)) {
    conv_name = "factory";
  } else {
    printf("Internal Error: Unknown domain call to \"%s\"\n", name);
    conv_name = "???";
  }

  result = string("m") + int2string(conv_name.length()) + conv_name + ending;

  if ((!isDomain) || genStatic) {
    if (isDomain) {
      result += string("mT") + int2string(int2string(arity).length() + 8) + "tiDomain" + int2string(arity) + "7domains2ti";
    } else {
      result += string("mT") + int2string(int2string(arity).length() + 12) + "tiRectDomain" + int2string(arity) + "7domains2ti";
    }
  }

  return result;
}

static const string TypeFieldAccessNode_emitMethodCall(TreeNode *n,
						       TypeNode *t,
						       CodeContext &context,
						       TreeListNode &params)
{
  Decl *d = n->simpName()->decl();

  string method = d->cMethodNameStatic();
  vector< string > parameters;
  params.emitExpressionList( context, parameters );

  if (t->isPointType())
    method = pointDispatch(d, t->tiArity());
  else if (t->isRectDomainType())
    method = domainDispatch(d, t->tiArity(), false, true);
  else if (t->isDomainType())
    method = domainDispatch(d, t->tiArity(), true, true);
  else if (t->isTitaniumArrayType()) {
    const TitaniumArrayTypeNode &node = *reinterpret_cast<TitaniumArrayTypeNode *>(t);
    method = string("_ti_arrayclassmethod_") + *(d->name())
	 + '(' + node.elementType()->cType()
 	 + ", " + decimalString(node.expr()->typeName()) + ')';
  } else if (t->isJavaArrayType())
    n->unimplemented( "method calls on Java arrays" );

  return locComment(n) + "/*TypeFAN*/ "
    + method + '(' + delimit( parameters ) + ')';
}

const string TypeFieldAccessNode::emitMethodCall( CodeContext &context,
						  TreeListNode &params )
{
  return TypeFieldAccessNode_emitMethodCall(this, ftype(),
					    context, params);
}


static const string SuperFieldAccessNode_emitMethodCall(TreeNode *n,
							CodeContext &context,
							TreeListNode &params)
{
    MethodDecl *d = static_cast< MethodDecl * >( n->simpName()->decl() );

    vector< string > parameters;
    pushThis( MANGLE_THIS_VAR(+), *d, parameters, *d->container()->asType(), true );
    params.emitExpressionList( context, parameters );

    const string method = d->cMethodNameStatic();
    return locComment(n) + "/*SFAN*/ " + method + '(' + delimit( parameters ) + ')';
}

const string ObjectFieldAccessNode::emitMethodCall( CodeContext &context, TreeListNode &params )
{
    MethodDecl *d = static_cast< MethodDecl * >( simpName()->decl() );
    TypeNode &t = *object()->type();

    if (d->modifiers() & Common::Static)
      return TypeFieldAccessNode_emitMethodCall(this, &t, context, params);

    if (isThisNode(object()) && t.decl() != d->container())
      return SuperFieldAccessNode_emitMethodCall(this, context, params);

    vector< string > parameters;
    const string instance = object()->defineTemporary( context );
    string method = dispatch( context, *this, instance, parameters );
    params.emitExpressionList( context, parameters );

    if (t.isPointType())
	method = pointDispatch(d, t.tiArity());
    else if (t.isRectDomainType()) {
      if (strcmp(d->name()->c_str(), "slice") == 0) {
	return rectDomainMethodSlice(d,
				     instance,
				     t.tiArity(),
				     parameters[1],
				     context,
				     getTemporary(),
				     getTemporary(),
				     getTemporary());
      } else {
	method = domainDispatch(d, t.tiArity(), false, false);
      }
    } else if (t.isDomainType()) {
      ClassDecl &classDecl = static_cast< ClassDecl & >(*t.decl());
      assert( classDecl.category() == Decl::Class );
      
      const string descTemp = getClassInfo( context, classDecl,
					    instance, t.isLocal() );
      method = descTemp + "->" + domainDispatch(d, t.tiArity(), true, false);
      
    } else if (t.isTitaniumArrayType()) {
      const TitaniumArrayTypeNode &node = reinterpret_cast<TitaniumArrayTypeNode &>(t);
      method = arrayMethodPrefix(&t) + *(d->name())
	+ '(' + node.elementType()->cType() + ", "
	+ decimalString(node.expr()->typeName()) + ')';
      
    } else if (t.isJavaArrayType())
	unimplemented( "method calls on java arrays" );

    return locComment(this) + "/*ObFAN*/ " + method + '(' + delimit( parameters ) + ')';
}

const string callGridMethod(TypeNode &t, const string &methodName,
			    const string &args)
{
  const TitaniumArrayTypeNode &node =
    reinterpret_cast<TitaniumArrayTypeNode &>(t);

  return arrayMethodPrefix(&t) + methodName
    + '(' + node.elementType()->cType() + ", "
    + decimalString(node.expr()->typeName()) + ')' + args;
}
